#!/usr/bin/perl
# combine predicted results

if( $#ARGV != 2 ){
	print "Usage: (0)predictions_list (1)agreement (2)output\n";
	print "       predictions_list is a list of the .predict filenames\n";
	print "       agreement is the number of files that must agree for the hit to be counted\n";
	print "       output is the output name (SVM scores will not be retained)\n";
	exit;
}

$list = `awk '{printf(\"\%s \",\$1)}' < $ARGV[0]`;
chomp($list);
@files = split(' ',$list);

if($ARGV[1] <= 0 || $ARGV[1] > $#files+1){
	print "Error: choose agreement to be between 1 and the length of the list\n";
	exit;
}

#
# determine format of files (number of columns)
# if 2 column, then <index> <prediction>
# if 4 columns, then <index> <seq> <prediction> <score>
# if 5 columns, then <index> <seq> <prediction> <score> <pval>
# if 7 columns, then <index> <chr> <start> <strand> <seq> <prediction> <score>
# if 8 columns, then <index> <chr> <start> <strand> <seq> <prediction> <score> <pval>
# else exit w error
#
open(IN,"$files[0]") || die "Error: can't open file $files[0]\n";
$line = <IN>;
chomp($line);
@parts = split(' ',$line);
if( $#parts == 1 ){$format = 1;}
elsif( $#parts == 3 ){$format = 2;}
elsif( $#parts == 4 ){$format = 3;}
elsif( $#parts == 7 ){$format = 4;}
elsif( $#parts == 8 ){$format = 5;}
else{
	print "Error: prediction file $files[0] an unexpected format\n";
	exit;
}
close(IN);

#
# read in all the files
#
@data = ();
@lens = ();
for($i=0; $i<= $#files; $i++){
	open(IN,"$files[$i]") || die "Error: can't open file $files[$i]\n";
	@fvals = ();
	while($line = <IN>){
		chomp($line);
		@parts = split(' ',$line);
		if( $format == 1 && $#parts != 1 ){
			print "Error: different format in file $files[$i]\n";
			exit;
		}
		if( $format == 2 && $#parts != 3 ){
			print "Error: different format in file $files[$i]\n";
			exit;
		}
		if( $format == 3 && $#parts != 4 ){
			print "Error: different format in file $files[$i]\n";
			exit;
		}
		if( $format == 4 && $#parts != 7 ){
			print "Error: different format in file $files[$i]\n";
			exit;
		}
		if( $format == 5 && $#parts != 8 ){
			print "Error: different format in file $files[$i]\n";
			exit;
		}
		push(@fvals,[@parts]);
	}
	close(IN);
	push(@lens, $#fvals);
	push(@data, [@fvals]);
}

#
# find max index
#
$max = $data[0][$lens[0]][0];
for($i=1; $i<= $#files; $i++){
	if( $max < $data[0][$lens[$i]][0] ){$max = $data[0][$lens[$i]][0];}
}

#
# search through all indices
#
@inds = ();
for($i=0; $i<= $#files; $i++){
	push(@inds,0);
}
$min = $data[0][$inds[0]][0];
for($i=1; $i<= $#files; $i++){
	if( $min > $data[$i][$inds[$i]][0] ){
		$min = $data[$i][$inds[$i]][0];
	}
}
open(OUT,">$ARGV[2]");
for($j=$min; $j<= $max; $j++){
	#
	# increment $j until it reaches the smallest current index
	#
	while($j< $min){$j++;}

	#
	# increment all indices until at least to $j
	#
	for($i=0; $i<= $#files; $i++){
		while( $data[$i][$inds[$i]][0] < $j ){$inds[$i]++;}
	}

	#
	# count matches
	#
	$match = 0;
	for($i=0; $i<= $#files; $i++){
		if( $data[$i][$inds[$i]][0] == $j ){
			$match++;
			$k=$i;
		}
	}

	#
	# if enough matches, print result
	#
	if( $match >= $ARGV[1] ){
		if( $format == 1 ){
			print OUT "$data[$k][$inds[$k]][0] $matche\n";
		}
		if( $format == 2 || $format == 3 ){
			print OUT "$data[$k][$inds[$k]][0] $data[$k][$inds[$k]][1] $match\n";
		}
		if( $format == 4 || $format == 5 ){
			print OUT "$data[$k][$inds[$k]][0] $data[$k][$inds[$k]][1] $data[$k][$inds[$k]][2] $data[$k][$inds[$k]][3] $data[$k][$inds[$k]][4] $match\n";
		}
	}

	#
	# update min
	#
	$min = $data[0][$inds[0]][0];
	for($i=1; $i<= $#files; $i++){
		if( $min > $data[$i][$inds[$i]][0] ){$min = $data[$i][$inds[$i]][0];}
	}
}
close(OUT);











