#!/usr/bin/perl
# parameterizations of sequences
# currently: requires $maxflank flanking sequences, can be upper or lower case

if( $#ARGV != 5 ){
	print "Usage: (0)input_sequences (1)output (2)flanking_length (3)params_list (4)scalars(or_-1_to_find) (5)stat_out\n";
	print "       input_sequences formatted as <sequence> <class> (class for training set only)\n";
	print "       4th input is list of max values for each feature, or -1 to determine these in the data\n";
	exit;
}

# write current status to status file, rather than stdout
open(STAT,">$ARGV[5]");
print STAT "Starting parameterization\n";

if( $ARGV[4] == -1 ){
	$findmax = 1;
}
else{
	$findmax = 0;
	@maxes = ();
	@maxl  = ();
	open( MAX,"$ARGV[4]" );
	while( $line = <MAX> ){
		chomp($line);
		@max = split(' ',$line);
		push( @maxes, [@max] );
		push( @maxl, $#max+1 );
	}
	close(MAX);
}

$maxflank = $ARGV[2]; # biggest flanking nts required; used to figure out the used length of input sequences
if( $maxflank < 0 ){
	print "Error in parameterize.pl: choose flanking length to be >=0\n";
	exit;
}

#
# read in sequences and check they are all the same length/no. cols, one per line
# fills @seqs array, plus $type and $seqlen values
#
&read_seqs();

#
# fill parameter array
# dimensions of allparams: ($#allparams+1) x ($#seqs+1) x ($seqlen) x ($numparams[$i])
#
# first get directory of list
@parts = split('/',$ARGV[3]);
if( $#parts > 0 ){
	$listdir = "$parts[0]";
	if( substr($ARGV[3],0,1) eq "/" ){
		$listdir = "";
	}
	for($i=1; $i< $#parts; $i++){
		$listdir = $listdir."/".$parts[$i];
	}
}
else{
	$listdir = ".";
}
open(LIST,"$ARGV[3]") || die "Error in parameterize.pl: Can't open params list $ARGV[3]\n";
@allparams = ();
@numparams = ();
@mods = ();
if( $findmax == 1 ){
	@maxes = ();
}
while( $line = <LIST> ){
	chomp($line);
	if( $line !~ /#/ ){
		$file = $listdir."/".$line;
		&read_param( $file );
		push( @allparams, [@par_array] );
		push( @numparams, $npars );
		push( @mods, $mod2 );
		if( $findmax == 1 ){
			push( @maxes, [@max] );
		}
	}
}
close(LIST);

if( $findmax == 0 ){
	if( $#maxes != $#mods ){
		print "Error in parameterize.pl: number of maxes in $ARGV[4] is not equal to the number of parameters\n";
		exit;
	}
	for($i=0; $i<= $#mods; $i++){
		if( $numparams[$i] != $maxl[$i] ){
			print "Error in parameterize.pl: number of max parameters in $ARGV[4], line $i is not equal to that of the input\n";
			exit;
		}
	}
}

#
# print parameter arrays in svm format
#
open(OUT,">$ARGV[1]");
for($j=0; $j<= $#seqs; $j++){
	$m=1;
	if( $type == 1 ){
		print OUT "$seqs[$j][1] ";
	}
	for($i=0; $i<= $#allparams; $i++){
		for($k=0; $k< $seqlen-$mods[$i]; $k++){
			for($l=0; $l< $numparams[$i]; $l++){
#				if( $allparams[$i][$j][$k][$l] != 0 ){
					printf OUT ("%i:%f ",$m,$allparams[$i][$j][$k][$l]/$maxes[$i][$l]);
#				}
				$m++;
			}
		}
	}
	print OUT "\n";
}
close(OUT);

# print out the maxes, if determined here
if( $findmax == 1 ){
	open(OUT,">$ARGV[1].maxes");
	for($i=0; $i<= $#allparams; $i++){
		for($l=0; $l< $numparams[$i]; $l++){
			printf OUT ("%f ",$maxes[$i][$l]);
		}
		print OUT "\n";
	}
	close(OUT);
}

print STAT "success\n";
close(STAT);

exit;


###################################################
#################### FUNCTIONS ####################
###################################################

#
# function to read the sequences (no input/output)
#
sub read_seqs{
	my $i;
	my $line;
	my @parts;
	open(SEQS,"$ARGV[0]") || die "Error in parameterize.pl: Can't open sequences file $ARGV[0]\n";
	@seqs = ();
	$type = -1;	# set to 1 if training data, 0 if test data
	# read in sequences
	while( $line = <SEQS> ){
		chomp($line);
		@parts = split(' ',$line);
		if( $type == -1 ){
			$type = $#parts;
			if( $type == 0 ){
#				print "   Parameterizing test data\n";
			}
			elsif( $type == 1 ){
#				print "   Parameterizing training data\n";
			}
			else{
				print "Error in parameterize.pl: unknown sequence format in $ARGV[0]\n";
				exit;
			}
		}
		else{
			if( $type != $#parts ){
				print "Error in parameterize.pl: inconsistent sequence format in $ARGV[0]\n";
				exit;
			}
		}
		push( @seqs, [@parts] );
	}
	close(SEQS);
	# check length
	$seqlen = 0;
	for($i=0; $i<= $#seqs; $i++){
		if( $seqlen == 0 ){
			$seqlen = length($seqs[$i][0])-2*$maxflank; # need $maxflank flanking sequences on each end
		}
		else{
			if( $seqlen != length($seqs[$i][0])-2*$maxflank ){
				print "Error in parameterize.pl: inconsistent sequence length in $ARGV[0]\n";
			}
		}
	}
}

#
# function to return index of a DNA n-mer
# ordering from AAA... to TTT..., incrementing A,C,G,T from the 1st col to the last
# USAGE: $_[0] = <n-mer>
#
sub seq_ind{
	my $i;
	my $len;
	my $index;
	$len = length($_[0]);
	$index = 0;
	for($i=0; $i< $len; $i++){
		$index+= &nt_ind( substr($_[0],$i,1) )*(4**$i);
	}
	return $index;
}

#
# function to translate a sequence of ACGT to 0123 (A/a=0, C/c=1, G/g=2, T/t=3)
# USAGE: $_[0] = A/C/G/T (upper or lower case)
#
sub nt_ind{
	if( $_[0] eq "A" || $_[0] eq "a" ){
		return 0;
	}
	elsif( $_[0] eq "C" || $_[0] eq "c" ){
		return 1;
	}
	elsif( $_[0] eq "G" || $_[0] eq "g" ){
		return 2;
	}
	elsif( $_[0] eq "T" || $_[0] eq "t" ){
		return 3;
	}
	else{
		print "Error in parameterize.pl: unknown nucleotide $_[0]\n";
		exit;
	}
}

#
# function to return the dna sequence starting at a given point, with flanking nts
# USAGE: $_[0] = <input sequence index> (0-$#seqs); $_[1] = <middle nt>; $_[2] = <# flanking seqs>;
#        $_[3] = <int>, goes an extra $_[3] nts to the right (for even-length sequences, usu. = 0 or 1)
#
sub choose_seq{
	if( $_[1]-$_[2] < 0 || $_[1]+$_[2]+$_[3] >= $seqlen+2*$maxflank ){
		print "Error in parameterize.pl: tried to access residues outside sequence length\n";
		exit;
	}
	return substr( $seqs[$_[0]][0], $_[1]-$_[2], 2*$_[2]+1+$_[3] );
}	

#
# read parameters from file <sequence> <<params>>
# USAGE: $_[0] = <filename>
# fills @par_array, $npars, $mod2, @max
# @par_array is ($#seqs+1) x ($seqlen) x ($npars), @max is ($npars)
#
sub read_param{
	my $i;
	my $j;
	my $k;
	my $line;
	my $sequence;
	my $flanks;
	my $seql;
	my @parts;
	my @parampts;
	my @fpparams;

	# read params, checking that
	#   - number of flanking sequences is <= flanking length
	#   - number of parameters and length of sequence on each line is the same
	#   - list is in the right order (of DNA sequences)
	if( $_[0] !~ /RANDOM/ ){
	open(PARS,"$_[0]") || die "Error in parameterize.pl: Can't open parameter file $_[0]\n";
	$npars = -1;
	$fpparams = ();
	$k=0;
	while( $line = <PARS> ){
		chomp($line);
		@parts = split(' ',$line);
		@parampts = ();
		for($i=1; $i<= $#parts; $i++){
			push(@parampts, $parts[$i]);
		}
		if( $npars == -1 ){
			$npars = $#parampts+1;
			if( $npars <= 0 ){
				print "Error in parameterize.pl: parameter file $_[0] has no parameters\n";
				exit;
			}
			$seql = length($parts[0]);
			$mod2 = ($seql+1)%2;
			if( $mod2 == 0 ){
				$flanks = ($seql-1)/2;
			}
			else{
				$flanks = ($seql-2)/2;
			}
			if( $flanks > $maxflank ){
				print "Error in parameterize.pl: parameter file $_[0] has sequences longer that suggested by your input flanking length\n";
				exit;
			}
		}
		else{
			if( $npars != $#parampts+1 || $seql != length($parts[0]) ){
				print "Error in parameterize.pl: parameter file $_[0] inconsistent\n";
				exit;
			}
		}
		if( $k != &seq_ind( $parts[0] ) ){
			print "Error in parameterize.pl: the sequences in parameter file $_[0] are not in order\n";
			exit;
		}
		push( @fpparams, [@parampts] );
		$k++;
	}
	close(PARS);
	
	# choose parameters
	@par_array = ();
	for($i=0; $i<= $#seqs; $i++){
		@seqpars = ();
#		if( $i%100000 == 0 ){
#			printf("     parameterizing %i of %i\n", $i, $#seqs+1);
#		}
		for($j=0; $j< $seqlen-$mod2; $j++){ #if even length sequence, stop just before the last residue
			$sequence = &choose_seq( $i, $maxflank+$j, $flanks, $mod2 );
			$k = &seq_ind( $sequence );
			if( $k> $#fpparams ){
				print "Error in parameterize.pl: tried to access larger index than total number of parameters in file $_[0]\n";
				exit;
			}
			push( @seqpars, $fpparams[$k] );
		}
		push( @par_array, [@seqpars] );
	}
	}
	# randomly assign parameters
	else{
	@parts = split('_',$_[0]);
	$npars = $parts[1];
	for($i=0; $i<= $#seqs; $i++){
		@seqpars = ();
		for($j=0; $j< $seqlen-$mod2; $j++){
			@parts = ();
			for($k=0; $k< $npars; $k++){
				push( @parts, rand() );;
			}
			push( @seqpars, [@parts] );
		}
		push( @par_array, [@seqpars] );
	}
	}
	
	# scale parameters so |biggest| is 1
	# will divide by $max when printing at the end; for some reason it screws stuff up here
	@max = ();
	for($k=0; $k< $npars; $k++){
		push( @max, 0 );
		for($i=0; $i<= $#seqs; $i++){
			for($j=0; $j< $seqlen-$mod2; $j++){
				if( $max[$k] < abs($par_array[$i][$j][$k]) ){
					$max[$k] = abs($par_array[$i][$j][$k]);
				}
			}
		}
	}
}





