#!/usr/bin/perl
# ranks features and selects the best

$p = 0.9;

if( $#ARGV != 1 ){
	print "Usage: input output\n";
	print "       Output lists the indices of the best features to use\n";
	exit;
}

#
# get max length of feature vectors
#
$lx = 0;
open(IN,"$ARGV[0]") || die "Can't open $ARGV[0]\n";
while($line = <IN>){
	chomp($line);
	@parts = split(' ', $line);
	($ind, $val) = split(':',$parts[$#parts]);
	if( $lx < $ind ){$lx=$ind;}
}
close(IN);

#
# read in training data (libSVM format)
#
@y = ();
@x = ();
open(IN,"$ARGV[0]") || die "Can't open $ARGV[0]\n";
while($line = <IN>){
	chomp($line);
	@parts = split(' ', $line);
	push(@y, $parts[0]);
	$j=1;
	@vec = ();
	for($i=1; $i<= $#parts; $i++){
		($ind, $val) = split(':',$parts[$i]);
		while($j< $ind){
			push(@vec, 0);
			$j++;
		}
		push(@vec, $val);
		$j++;
	}
	while($j<= $lx){
		push(@vec, 0);
		$j++;
	}
	push(@x, [@vec]);
	if( $#vec != $lx-1 ){
		print "Error in code: length of vector is not what it's supposed to be\n";
		exit;
	}
}
close(IN);
$nx = $#x+1;

#
# prepare array of feature dimension importances
#
@dimv = ();
for($i=0; $i< $lx; $i++){
	push(@dimv, [0,$i]);
}

#
# different functions to fill dimv
#
#&correlations();
&mi();

#
# pick best dimensions:
#    add up absolute values and take top 90%
#

# sort values, count sum (max)
@sorted = sort{ $a->[0] <=> $b->[0] } @dimv;
$max = 0;
for($i=0; $i< $lx; $i++){
	$max+= $sorted[$i][0];
}
$end = $p*$max;

# keep biggest values up to p*max
@shortened = ();
$i=$lx-1;
$sum=0;
while($sum<= $end){
	push(@shortened, [($sorted[$i][0], $sorted[$i][1])] );
	$i--;
	$sum+= $sorted[$i][0];
}

# resort by index to print
@keep = sort{ $a->[1] <=> $b->[1] } @shortened;

open(OUT,">$ARGV[1]");
for($i=0; $i<= $#shortened; $i++){
	print OUT "$keep[$i][1] $keep[$i][0]\n";
}
close(OUT);


exit;

# fill dimv with the correlation between each dimension and y
sub correlations{
	my $i;
	my $j;

	$meany = 0;
	$vary = 0;
	for($j=0; $j< $nx; $j++){
		$meany+= $y[$j];
	}
	$meany/= $nx;
	for($j=0; $j< $nx; $j++){
		$vary+= ($y[$j] - $meany)**2;
	}
	$vary= sqrt($vary);

	for($i=0; $i< $lx; $i++){
		$meanx = 0;
		$varx = 0;
		$corr = 0;
		for($j=0; $j< $nx; $j++){
			$meanx+= $x[$j][$i];
		}
		$meanx/= $nx;
		for($j=0; $j< $nx; $j++){
			$varx+= ($x[$j][$i] - $meanx)**2;
			$corr+= ($x[$j][$i] - $meanx)*($y[$j]-$meany);
		}
		$varx = sqrt($varx);
		$corr/= ($varx*$vary);
		$dimv[$i][0] = abs($corr);
	}
}

# fill dimv with mutual information between each dimension and y
# assumes each dimension is between -1 and 1, histograms in intervals of 0.1, and y is -1 or 1
# so hist of y has 2 bins, hist of x is 20 bins, hist of x-y is 40 bins
sub mi{
	my $i;
	my $j;
	my $k;
	my $l;
	$step = 0.1;
	$ybins = 2;
	$xbins = 2/$step;

	for($k=0; $k< $ybins; $k++){
		push(@hy, 0);
	}
	for($l=0; $l< $xbins; $l++){
		push(@hx, 0);
	}
	for($k=0; $k< $ybins; $k++){
		@tmp = ();
		for($l=0; $l< $xbins; $l++){
			push(@tmp, 0);
		}
		push(@hxy, [@tmp] );
	}

	# histogram of y
	for($j=0; $j< $nx; $j++){
		if( $y[$j] == 1 ){$hy[1]++;}
		elsif( $y[$j] == -1 ){$hy[0]++;}
		else{
			print "Error: $y[$j] is not +/- 1\n";
			exit;
		}
	}
	for($k=0; $k< $ybins; $k++){
		$hy[$k]/= $nx;
	}

	# now loop over each dimension
	for($i=0; $i< $lx; $i++){
		# histogram of x
		for($j=0; $j< $nx; $j++){
			$bin = sprintf("%i", ($x[$j][$i]+1)/$step );
			if( $x[$j][$i] == 1.0 ){$bin = $xbins-1;}
			if( $bin< 0 || $bin >= $xbins ){
				print "Error: $x[$j][$i] is not between +/- 1\n";
				exit;
			}
			$hx[$bin]++;
		}
		for($l=0; $l< $xbins; $l++){
			$hx[$l]/= $nx;
		}

		# histogram of x and y
		for($j=0; $j< $nx; $j++){
			$bin = sprintf("%i", ($x[$j][$i]+1)/$step );
			if( $x[$j][$i] == 1.0 ){$bin = $xbins-1;}
			if( $y[$j] == 1 ){$hxy[1][$bin]++;}
			else{ $hxy[0][$bin]++;}
		}
		for($k=0; $k< $ybins; $k++){
			for($l=0; $l< $xbins; $l++){
				$hxy[$k][$l]/= $nx;
			}
		}

		# compute mi
		$info = 0;
		for($k=0; $k< $ybins; $k++){
			for($l=0; $l< $xbins; $l++){
				if( $hxy[$k][$l] > 0 ){
					if( $hy[$k] == 0 || $hx[$l] == 0 ){
						print "Error: somehow the marginal prob is zero, but joint is non-zero\n";
						exit;
					}
					$info+= $hxy[$k][$l]*log($hxy[$k][$l]/($hy[$k]*$hx[$l]));
				}
			}
		}
		$dimv[$i][0] = abs($info);
	}
}






